/* Copyright (C) 1998 Odysseas Pentakalos and Aram Khalili

  This program is freely distributable under the GNU General
  Public License Version 2 or later. The GNU GPL is available
  from the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA 02111-1307, USA.

  The authors can be contacted by email :
  Odysseas Pentakalos : odysseas@cs.umbc.edu
  Aram Khalili        : akhali1@gl.umbc.edu
*/

#include <stdio.h>
#include <string.h>
#include <scsi/sg.h>
#include <scsi/scsi.h>
#include <sys/types.h>
#include "tapelib.h"

int mode_sense(int fd, __u16 pagecontrol, __u16 pagecode, elt_addr_assgn_t *eas,
	       tgd_t *tgd, dev_cap_t *dcp, LCD_t *LCD, parity_t *par)
{
  /* structure for the MODE SENSE SCSI command (6-byte command) */

  typedef struct {
    struct sg_header head;
    u_char cmd[6];
  } SCSI_Cmd_t;
  
  /* function pointers */
  
  void elt_addr_assgn(char *, elt_addr_assgn_t *);
  void trans_geo_desc(char *, tgd_t *);
  void dev_cap_p(char *, dev_cap_t *);
  void LCD_p(char *, LCD_t *);
  void parity(char *, parity_t *);
  
  SCSI_Cmd_t Cmd;
  int i, ret;
  unsigned char *buf, *offset;

  /* blank command */

  for (i=0; i<6; i++)
    Cmd.cmd[i]=0;

  Cmd.cmd[0]=MODE_SENSE; /* set command */
  Cmd.cmd[1]=0x08; /* LUN = 0, reserved fields = 0, DBD=1 */

  /* Now set pagecontrol (bits 7 & 8 Cmd.cmd[2]) :
     
     00 : Current values
     01 : changeable values (returns bitmask: changeable=1, unchangeable=0)
     10 : (factory-)default values
     11 : saved (to non-volatile memory) values

   */

  switch (pagecontrol) 
    {
    case 0:  Cmd.cmd[2]=0; break;
    case 1:  Cmd.cmd[2]=64; break;
    case 2:  Cmd.cmd[2]=128; break;
    case 3:  Cmd.cmd[2]=192; break;
    default: /* invalid */
             fprintf (stderr, "MODE_SENSE: Invalid pagecontrol\n");
	     return 0;
    }
      
#ifdef DEBUG
  printf("pagecontrol = %d, cmd[2] = %d\n", pagecontrol, Cmd.cmd[2]);
#endif

  /* Setting allocation lengths; 4 header bytes are added */

  switch (pagecode)
    {
    case 0: Cmd.cmd[4]=8; break;
    case 0x1D: Cmd.cmd[4]=24; break;
    case 0x1E: Cmd.cmd[4]=8; break;
    case 0x1F: Cmd.cmd[4]=24; break;
    case 0x22: Cmd.cmd[4]=88; break;
    case 0x3F: Cmd.cmd[4]=136; break;
    default: /* invalid */
             fprintf(stderr,"MODE_SENSE: Invalid pagecode(%d)\n", pagecontrol);
	     return 0;
    }
  Cmd.cmd[2]|=pagecode;

#ifdef DEBUG
  printf("pagecode = %d, cmd[2] = %d\n", pagecode, Cmd.cmd[2]);
#endif

  Cmd.cmd[3]=0;
  Cmd.cmd[5]=0;

  /* Prepare the SCSI command header */
  Cmd.head.pack_len=sizeof(SCSI_Cmd_t);
  Cmd.head.reply_len=Cmd.cmd[4]+sizeof(struct sg_header);

#ifdef DEBUG
  printf("mode_sense: Sending out the MODE_SENSE command\n");
  printf("mode sense: using pagecode %d, pagecontrol %d.\n", pagecode, pagecontrol);
#endif DEBUG

  ret=write(fd, &Cmd, sizeof(SCSI_Cmd_t));

#ifdef DEBUG
  printf("mode_sense: The write returned %d status\n", ret);
#endif DEBUG

  /* malloc return buffer */

  buf=(u_char*)malloc(Cmd.head.reply_len+1);

  /* Now lets read back the results */

  ret=read(fd, buf, Cmd.head.reply_len+1);

#ifdef DEBUG
  printf("move_medium: The read returned %d bytes (expected %d)\n", ret, Cmd.head.reply_len);
#endif DEBUG

  /* Skip the struct sg_header header */

  offset = buf + sizeof(struct sg_header) + 4;
  
  for (i=0;i<Cmd.head.reply_len;i++)
    printf("%d\n",offset[i]);


  /* find out which information to look for */

  switch(pagecode)
    {
    case 0: parity(offset, par); break;
    case 0x1D: elt_addr_assgn(offset,eas); break;
    case 0x1E: trans_geo_desc(offset,tgd); break;
    case 0x1F: dev_cap_p(offset, dcp); break;
    case 0x22: LCD_p(offset, LCD); break;
    case 0x3F: elt_addr_assgn(offset, eas);trans_geo_desc(offset+20, tgd);dev_cap_p(offset+24, dcp);LCD_p(offset+44, LCD);parity(offset+128, par); break;
    default: /* invalid */ fprintf(stderr,"MODE_SENSE: invalid pagecode\n"); return 0;
    }

  return ret;

}

/* Function: void elt_addr_assgn(char *buf, elt_addr_assgn_t *eas)

   arguments: buf: a character pointer to the start of the element address 
                   assigment data
	      eas: a elt_addr_assgn_t pointer to a structure that will hold
	           the requested sense data upon return

   returns: nothing
*/

void elt_addr_assgn(char *buf, elt_addr_assgn_t *eas)
{
  eas->ps=(buf[0] & 0x80) ? 1 : 0;
  eas->tea = (buf[2]<<8)|buf[3];
  eas->num_trans_elt = (buf[4]<<8)|buf[5];
  eas->sea = (buf[6]<<8)|buf[7];
  eas->num_stor_elt = (buf[8]<<8)|buf[9];
  eas->iea = (buf[10]<<8)|buf[11];
  eas->num_imp_elt = (buf[12]<<8)|buf[13];
  eas->dea = (buf[14]<<8)|buf[15];
  eas->num_data_elt = (buf[16]<<8)|buf[17];
}

/* Function: void trans_geo_desc(char *buf, tgd_t *tgd)

   arguments: buf: a character pointer to the start of the transport geometry data
	      tgd: a tgd_t pointer to a structure that will hold the requested
	           sense data upon return

   returns: nothing
*/

void trans_geo_desc(char *buf, tgd_t *tgd)
{
  tgd->ps = (buf[0] & 0x80) ? 1 : 0;
  tgd->rot = (buf[2] & 0x01)? 1 : 0;
  tgd->trans_elt_no = buf[3];
}

/* Function: void dev_cap_p(char *buf, dev_cap_t *dcp)

   arguments: buf: a character pointer to the start of the transport geometry data
	      dcp: a dev_cap__t pointer to a structure that will hold the requested
	           sense data upon return

   returns: nothing
*/

void dev_cap_p(char *buf, dev_cap_t *dcp)
{
  dcp->ps = (buf[0] & 0x80) ? 1 : 0;

  /* some data returned doesn't seem right; it's not what the manual
     says should be returned */

  dcp->mt = ((buf[2] & 0x01)<<4)|(buf[4]);
  dcp->st = ((buf[2] & 0x02)<<3)|(buf[5]);
  dcp->ie = ((buf[2] & 0x04)<<2)|(buf[6]);
  dcp->dt = ((buf[2] & 0x08)<<1)|(buf[7]);
  /* printf("%d %d %d %d %d\n", buf[2], buf[4], buf[5], buf[6], buf[7]); */

}

/* Function: void LCD_p(char *buf, LCD_t *LCD)

   arguments: buf: a character pointer to the start of the LCD data
	      LCD: a LCD_t pointer to a structure that will hold the requested
	           sense data upon return

   returns: nothing
*/

void LCD_p(char *buf, LCD_t *LCD)
{ 
  int x;

  LCD->ps = (buf[0] & 0x80) ? 1 : 0;
  LCD->LCD_sec = (buf[2] & 0x40) ? 1 : 0;
  LCD->wl1 = (buf[2] & 0x8) ? 1 : 0;
  LCD->wl2 = (buf[2] & 0x4) ? 1 : 0;
  LCD->wl3 = (buf[2] & 0x2) ? 1 : 0;
  LCD->wl4 = (buf[2] & 0x1) ? 1 : 0;
  
  for(x=0;x<20;x++)
    LCD->line1[x]=buf[x+4];

  for(x=0;x<20;x++)
    LCD->line2[x]=buf[x+24];

  for(x=0;x<20;x++)
    LCD->line3[x]=buf[x+44];

  for(x=0;x<20;x++)
    LCD->line4[20]=buf[x+64];
}

/* Function: void parity(char *buf, parity_t *par)

   arguments: buf: a character pointer to the start of the parity data
	      par: a parity_t_t pointer to a structure that will hold the requested
	           sense data upon return
	     
   returns: nothing
*/

void parity(char *buf, parity_t *par)
{
  par->ps = (buf[0] & 0x80) ? 1 : 0;
  par->par = (buf[2] & 0x20) ? 1 : 0;
  par->retries = buf[3];
}
